##############################################################################
#
# Copyright (c) 2002-2015 Nexedi SA and Contributors. All Rights Reserved.
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
#
##############################################################################
from cStringIO import StringIO
from httplib import NO_CONTENT
import json
import msgpack
import string
import random

from Products.ERP5Type.tests.SecurityTestCase import SecurityTestCase


def generateRandomString():
  return ''.join([random.choice(string.ascii_letters + string.digits) \
    for _ in xrange(32)])

class ORSWendelinTest(SecurityTestCase):
  """
  ORS Wendelin Test
  """

  def getTitle(self):
    return "ORS Wendelin Test"

  def afterSetUp(self):
    # Set up variables for ORS ingestion testing
    self.ors_ingestion = self.portal.portal_ingestion_policies.ors_ingestion
    self.test_ors_example_log_valid = {'log': self.portal.web_page_module.test_ors_example_log_valid.getTextContent()}
    self.test_ors_example_log_invalid = {'log': self.portal.web_page_module.test_ors_example_log_invalid.getTextContent()}

  def beforeTearDown(self):
    self.abort()

    # Clean up all test objects
    for module, portal_type in (
      (self.portal.project_module, 'Project'),
      (self.portal.person_module, 'Person'),
      (self.portal.data_acquisition_unit_module, 'Data Acquisition Unit'),
      (self.portal.data_supply_module, 'Data Supply'),
      (self.portal.data_stream_module, 'Data Stream'),
      (self.portal.data_ingestion_module, 'Data Ingestion'),
      (self.portal.data_analysis_module, 'Data Analysis'),
      (self.portal.data_array_module, 'Data Array'),
    ):
      objects = module.objectValues(portal_type=portal_type)
      if objects:
        test_object_ids = [obj.getId() for obj in objects if ('test' in obj.getReference() and 'default' not in obj.getId())]
        if test_object_ids:
          module.manage_delObjects(ids=test_object_ids)
    self.tic()

  def _createOrsWendelinUser(self, reference, project, role):
    # Create and validate a new Person with an assignment associated to the given project and the given role
    user = self.portal.person_module.newContent(portal_type='Person',
      reference=reference)
    user.newContent(portal_type='Assignment', destination_project=project, role=role).open()
    user.newContent(portal_type='ERP5 Login', reference=reference, password=reference).validate()
    user.validate()
    self.tic()
    return user

  def _createOrsClientProject(self, reference=None, client_user_reference=None, ors_reference=None, ors_title_reference=None):
    # Create a client project with the given reference. The reference will be applied to all items linked to the project
    # If set, client_user_reference overrides the reference of the client user account
    # If set, ors_reference overrides the reference of the Data Acquisition Unit (ORS)
    # If set, ors_title_reference overrides the title reference of the Data Acquisition Unit
    if reference is None:
      reference = generateRandomString()
    references_dict = {
      'project': 'test_project_%s' % reference,
      'client_user': 'test_user_%s' % (client_user_reference or reference),
      'generic': 'test_%s' % (ors_reference or reference),
      'ors': 'ors.test_%s' % reference,
      'title': 'ORS Test %s' % (ors_title_reference or reference)
    }

    # Call the script responsible for creating the project and all associated items and store the JSON response
    response = self.portal.ERP5Site_registerOrsClientProject(
      references_dict['project'],
      references_dict['project'],
      references_dict['client_user'],
      references_dict['generic'],
      references_dict['title']
    )
    self.tic()

    # Fetch all associated items from the catalog
    project = self.portal.portal_catalog.getResultValue(portal_type='Project',
      reference=references_dict['project'],
      validation_state='validated')

    client_user = self.portal.portal_catalog.getResultValue(portal_type='Person',
      reference=references_dict['client_user'],
      validation_state='validated')

    data_acquisition_unit = self.portal.portal_catalog.getResultValue(portal_type='Data Acquisition Unit',
      reference=references_dict['ors'],
      title=references_dict['title'],
      validation_state='validated')

    data_supply = None
    if data_acquisition_unit:
      data_supply = data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)

    # Return all associated objects
    return {
      'response': response,
      'project': project,
      'client_user': client_user,
      'data_acquisition_unit': data_acquisition_unit,
      'data_supply': data_supply
    }

  def _ingestOrsDataFromFluentd(self, log_data, reference):
    # Simulate a fluentd instance sending the given log data to Wendelin for ingestion
    body = msgpack.packb([0, log_data], use_bin_type=True)
    env = {'CONTENT_TYPE': 'application/octet-stream'}
    path = self.ors_ingestion.getPath() + '/ingest?reference=' + reference
    publish_kw = dict(user='ERP5TypeTestCase', env=env,
      request_method='POST', stdin=StringIO(body))
    return self.publish(path, **publish_kw)

  def _getDataStream(self, data_acquisition_unit):
    # Get a Data Stream linked to the given Data Acquisition Unit
    for line in data_acquisition_unit.getAggregateRelatedValueList(portal_type='Data Ingestion Line'):
      data_stream = line.getAggregateValue(portal_type='Data Stream')
      if data_stream:
        return data_stream

  def _getDataAnalysis(self, data_supply):
    # Get a Data Analysis linked to the given Data Supply
    for data_analysis in data_supply.getSpecialiseRelatedValueList(portal_type='Data Analysis'):
      if data_analysis:
        return data_analysis

  def _getDataArrays(self, data_analysis):
    # Get a Data Array linked to the given Data Analysis
    data_arrays = []
    for line in data_analysis.contentValues(portal_type='Data Analysis Line'):
      data_array = line.getAggregateValue(portal_type='Data Array')
      if data_array:
        data_arrays.append(data_array)
    return data_arrays

  def _getOrsIngestionObjects(self, log_data, reference, stop_data_analysis=False):
    # Simulate an ingestion of the given log data
    response = self._ingestOrsDataFromFluentd(log_data, reference)
    self.tic()

    # Retrieve all items linked to the ingestion
    data_acquisition_unit = self.portal.portal_catalog.getResultValue(portal_type='Data Acquisition Unit', reference=reference)
    data_supply = data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)
    self.tic()

    data_stream = self._getDataStream(data_acquisition_unit)

    # Manually call the data analysis alarms to process the data into KPIs ASAP
    self.portal.portal_alarms.wendelin_handle_analysis.activeSense()
    self.tic()
    self.portal.portal_alarms.wendelin_handle_analysis.activeSense()
    self.tic()

    data_analysis = self._getDataAnalysis(data_supply)
    if stop_data_analysis and data_analysis.getSimulationState() == 'started':
      data_analysis.stop()
      self.tic()

    data_arrays = self._getDataArrays(data_analysis)

    # Return all associated objects
    return {
      'response': response,
      'data_acquisition_unit': data_acquisition_unit,
      'data_supply': data_supply,
      'data_stream': data_stream,
      'data_analysis': data_analysis,
      'data_arrays': data_arrays
    }

  def _checkDocumentPermissions(self, user, document, user_can_view, user_can_modify, user_can_add):
    user_id = user.Person_getUserId()

    # Check if the given user can access and view the document
    if user_can_view:
      self.assertUserCanAccessDocument(user_id, document)
      self.assertUserCanViewDocument(user_id, document)
    else:
      self.failIfUserCanAccessDocument(user_id, document)
      self.failIfUserCanViewDocument(user_id, document)

    # Check if the given user can edit the document
    if user_can_modify:
      self.assertUserCanModifyDocument(user_id, document)
    else:
      self.failIfUserCanModifyDocument(user_id, document)

    # Check if the given user can add a document
    if user_can_add:
      self.assertUserCanAddDocument(user_id, document)
    else:
      self.failIfUserCanAddDocument(user_id, document)

  def _checkIngestionDocumentsPermissions(self, user, ingestion_objects, user_is_manager, same_project):
    # A client can only view a Data Acquisition Unit (ORS) if it is associated to their project
    # A manager can view and edit all of them, as well as add one
    self._checkDocumentPermissions(user, ingestion_objects['data_acquisition_unit'], user_is_manager or same_project, user_is_manager, user_is_manager)

    # Same as above for a Data Supply (required for scoping clients to their projects)
    self._checkDocumentPermissions(user, ingestion_objects['data_supply'], user_is_manager or same_project, user_is_manager, user_is_manager)

    # Only a manager can view a Data Stream, and nothing else
    self._checkDocumentPermissions(user, ingestion_objects['data_stream'], user_is_manager, False, False)

    # A client can view a Data Array if it is associated to their project
    # A manager can view all of them
    for data_array in ingestion_objects['data_arrays']:
      self._checkDocumentPermissions(user, data_array, user_is_manager or same_project, False, False)

  def _checkModulePermissions(self, user, user_is_manager):
    # Everyone can view the Data Acquisition Unit and Data Supply modules
    # Only managers can add documents to them
    self._checkDocumentPermissions(user, self.portal.data_acquisition_unit_module, True, False, user_is_manager)
    self._checkDocumentPermissions(user, self.portal.data_supply_module, True, False, user_is_manager)

    # Only managers can view the Data Ingestion and Data Stream modules
    self._checkDocumentPermissions(user, self.portal.data_ingestion_module, user_is_manager, False, False)
    self._checkDocumentPermissions(user, self.portal.data_stream_module, user_is_manager, False, False)

    # Everyone can view the Data Transformation module (required for KPI graphing)
    self._checkDocumentPermissions(user, self.portal.data_transformation_module, True, False, False)

    # Only managers can view the Data Analysis module
    self._checkDocumentPermissions(user, self.portal.data_analysis_module, user_is_manager, False, False)

    # Everyone can view the Data Array module (required for KPI graphing)
    self._checkDocumentPermissions(user, self.portal.data_array_module, True, False, False)

    # Only managers have access to the Person and Project modules and can add items to them for client management purposes
    self._checkDocumentPermissions(user, self.portal.project_module, user_is_manager, False, user_is_manager)
    self._checkDocumentPermissions(user, self.portal.person_module, user_is_manager, False, user_is_manager)

  def test_01_createOrsDataSupply(self):
    '''
    Test the action which creates an ORS Data Supply from a Data Acquisition Unit.
    Check that the Data Supply is indeed created and validated.
    '''
    ors_reference = 'ors.test_%s' % generateRandomString()

    # Create and validate a Data Acquisition Unit
    data_acquisition_unit = self.portal.data_acquisition_unit_module.newContent(
      portal_type='Data Acquisition Unit',
      reference=ors_reference
    )
    data_acquisition_unit.validate()
    self.tic()

    # Call the tested script which creates a related Data Supply
    data_supply = data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)
    self.tic()

    # Check that the Data Supply exists and is validated
    self.assertTrue(data_supply is not None)
    self.assertTrue(data_supply.getValidationState() == 'validated')

  def test_02_registerOrsClientProject(self):
    '''
    Test the script called during slave instantiation in SlapOS to register a new client project.
    Check the successful case as well as all error cases.
    '''

    # Generate a random reference, call the tested script and retrieve all associated objects
    # This first call should succeed
    reference = generateRandomString()
    project_objects = self._createOrsClientProject(reference=reference)

    # Parse the JSON response and check that the user credentials exist and are valid
    response_dict = json.loads(project_objects['response'])
    self.assertTrue(response_dict['username'] == 'test_user_%s' % reference)
    self.assertTrue(len(response_dict['init_password']) == 16)

    # Check that all objects created by the script have been created
    self.assertTrue(project_objects['project'] is not None)
    self.assertTrue(project_objects['client_user'] is not None)
    self.assertTrue(project_objects['data_acquisition_unit'] is not None)
    self.assertTrue(project_objects['data_supply'] is not None)

    # Call the script a second time with the same reference
    # This should not do anything and respond with an error as the project already exists
    project_objects = self._createOrsClientProject(reference=reference)

    # Parse the JSON response and check that the error message is valid
    response_dict = json.loads(project_objects['response'])
    self.assertTrue('error_msg' in response_dict)
    self.assertTrue(response_dict['error_msg'] == "Client project %s already registered." % project_objects['project'].getReference())

    # Create a new reference for the project, but reuse the previous reference for the client user account
    new_project_reference = generateRandomString()
    while new_project_reference == reference:
      new_project_reference = generateRandomString()
    # Call the script a third time, keeping the same reference as before ONLY for the client user account
    # This should also error out as the client user account already exists
    project_objects = self._createOrsClientProject(reference=new_project_reference, client_user_reference=reference)

    # Parse the JSON response and check that the error message is valid
    response_dict = json.loads(project_objects['response'])
    self.assertTrue('error_msg' in response_dict)
    self.assertTrue(response_dict['error_msg'] == "Client account username not available: test_user_%s." % reference)

    # Check that the new project is NOT created
    self.assertTrue(project_objects['project'] is None)

    # Same thing as above, but reuse the original reference for the Data Acquisition Unit (ORS)
    new_project_reference_2 = generateRandomString()
    while (new_project_reference_2 == reference) or (new_project_reference_2 == new_project_reference):
      new_project_reference_2 = generateRandomString()
    # Call the script a fourth time, keeping the same reference as before ONLY for the Data Acquisition Unit
    # This should also error out as it already exists
    project_objects = self._createOrsClientProject(reference=new_project_reference_2, ors_reference=reference)

    # Parse the JSON response and check that the error message is valid
    response_dict = json.loads(project_objects['response'])
    self.assertTrue('error_msg' in response_dict)
    self.assertTrue(response_dict['error_msg'] == "ORS with tag test_%s already registered." % reference)

    # Check that the new project is NOT created
    self.assertTrue(project_objects['project'] is None)

    # Generate new reference again, but reuse the original reference for the Data Acquisition Unit's title
    new_project_reference_3 = generateRandomString()
    while (new_project_reference_3 == reference):
      new_project_reference_3 = generateRandomString()
    # Call the script a fifth time, using the new reference for everything except the ORS title
    # This should succeed as the title is not checked for collision
    project_objects = self._createOrsClientProject(reference=new_project_reference_3, ors_title_reference=reference)

    # Parse the JSON response and check that the user credentials exist and are valid
    response_dict = json.loads(project_objects['response'])
    self.assertTrue(response_dict['username'] == 'test_user_%s' % new_project_reference_3)
    self.assertTrue(len(response_dict['init_password']) == 16)

    # Check that all objects created by the script have been created
    self.assertTrue(project_objects['project'] is not None)
    self.assertTrue(project_objects['client_user'] is not None)
    self.assertTrue(project_objects['data_acquisition_unit'] is not None)
    self.assertTrue(project_objects['data_supply'] is not None)

  def test_03_registerNewOrsToProject(self):
    '''
    Test the action which configures a Data Acquisition Unit and Data Supply representing an ORS
    assigned to a given client project.
    '''

    project_objects = self._createOrsClientProject()

    # Check that all objects related to the project exist
    self.assertTrue(project_objects['project'] is not None)
    self.assertTrue(project_objects['client_user'] is not None)
    self.assertTrue(project_objects['data_acquisition_unit'] is not None)
    self.assertTrue(project_objects['data_supply'] is not None)

    # Add a new ORS to the project, with a new reference
    reference = generateRandomString()
    generic_reference = 'test_%s' % reference
    ors_reference = 'ors.test_%s' % reference
    ors_title = 'ORS Test %s' % reference

    project_objects['project'].Project_newOrs(generic_reference, ors_title)
    self.tic()

    # Check that the Data Acquisition Unit exists and is validated
    data_acquisition_unit = self.portal.portal_catalog.getResultValue(portal_type='Data Acquisition Unit',
      reference=ors_reference,
      title=ors_title,
      validation_state='validated')
    self.assertTrue(data_acquisition_unit is not None)

    # Check that the Data Supply exists and is validated
    data_supply = data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)
    self.assertTrue(data_supply is not None)

    # Generate new references
    new_reference = generateRandomString()
    new_generic_reference = 'test_%s' % new_reference
    new_ors_reference = 'ors.test_%s' % new_reference
    new_ors_title = 'ORS Test %s' % new_reference

    # Add a new ORS to the project, with a different reference but the same title
    project_objects['project'].Project_newOrs(new_generic_reference, ors_title)
    self.tic()

    # Check that the Data Acquisition Unit exists and is validated
    new_data_acquisition_unit = self.portal.portal_catalog.getResultValue(portal_type='Data Acquisition Unit',
      reference=new_ors_reference,
      title=ors_title,
      validation_state='validated')
    self.assertTrue(new_data_acquisition_unit is not None)

    # Check that the Data Supply exists and is validated
    new_data_supply = new_data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)
    self.assertTrue(new_data_supply is not None)

    # Add another ORS to the project, with the same reference and a new title
    # This should fail as there can be no reference collision
    project_objects['project'].Project_newOrs(generic_reference, new_ors_title)
    self.tic()

    # Check that the Data Acquisition Unit does not exist
    new_data_acquisition_unit = self.portal.portal_catalog.getResultValue(portal_type='Data Acquisition Unit',
      reference=ors_reference,
      title=new_ors_title,
      validation_state='validated')
    self.assertTrue(new_data_acquisition_unit is None)

  def test_04_1_ingestValidOrsLogDataFromFluentd(self, valid_data=True):
    '''
    Test a simple valid ORS log ingestion: simulate a fluentd gateway forwarding valid ORS logs to a project on the platform,
    and check that all objects related to the ingestion are valid.
    '''
    project_objects = self._createOrsClientProject()
    ors_reference = project_objects['data_acquisition_unit'].getReference()

    # Get the correct data logs according to what is being tested
    test_ors_example_log = self.test_ors_example_log_valid
    if not valid_data:
      test_ors_example_log = self.test_ors_example_log_invalid
    ingestion_objects = self._getOrsIngestionObjects(test_ors_example_log, ors_reference)

    # In all cases, check that all objects related to the ingestion exist
    self.assertEqual(NO_CONTENT, ingestion_objects['response'].getStatus())
    self.assertTrue(ingestion_objects['data_acquisition_unit'] is not None)
    self.assertTrue(ingestion_objects['data_supply'] is not None)
    self.assertTrue(ingestion_objects['data_stream'].getData() is not None)
    self.assertTrue(ingestion_objects['data_analysis'] is not None)
    self.assertTrue(all(data_array is not None for data_array in ingestion_objects['data_arrays']))

    # Check that the data arrays containing the KPI data have correctly been initialized
    self.assertTrue(ingestion_objects['data_acquisition_unit'].Base_getERABDataArrayKey() is not None)
    self.assertTrue(ingestion_objects['data_acquisition_unit'].Base_getEUTRANDataArrayKey() is not None)

    if valid_data:
      # If all data is valid: check that the arrays contain some data
      self.assertTrue(all(data_array.getArrayShape() is not None for data_array in ingestion_objects['data_arrays']))
      self.assertTrue(all(data_array.getArrayDtype() is not None for data_array in ingestion_objects['data_arrays']))
    else:
      # If some data is invalid: check that the whole log (one data chunk in this case) is ignored
      # and the data arrays are empty
      self.assertTrue(all(data_array.getArrayShape() is None for data_array in ingestion_objects['data_arrays']))
      self.assertTrue(all(data_array.getArrayDtype() is None for data_array in ingestion_objects['data_arrays']))
  
  def test_04_2_ingestInvalidOrsLogDataFromFluentd(self):
    '''
    Test an invalid ORS log ingestion: simulate a fluentd gateway forwarding invalid ORS logs to a project on the platform.
    Check that all objects are valid, but that the data arrays contain no data due to the invalid data chunk being ignored.
    '''

    # Call the above test, but test with invalid data
    self.test_04_1_ingestValidOrsLogDataFromFluentd(valid_data=False)

  def test_05_orsWendelinSecurityModel(self):
    '''
    Test ORS Wendelin's custom security model: check that different users can only access the ingestion documents relative to their role and project.
    '''

    # Create two distinct projects
    project_A_objects = self._createOrsClientProject()
    ors_A_reference = project_A_objects['data_acquisition_unit'].getReference()
    project_B_objects = self._createOrsClientProject()
    ors_B_reference = project_B_objects['data_acquisition_unit'].getReference()

    # Perform ingestions for the ORSs of both projects
    ingestion_A_objects = self._getOrsIngestionObjects(self.test_ors_example_log_valid, ors_A_reference, stop_data_analysis=True)
    ingestion_B_objects = self._getOrsIngestionObjects(self.test_ors_example_log_valid, ors_B_reference, stop_data_analysis=True)

    # Create a client user not associated to a project (should not happen in practice)
    client_user_N = self._createOrsWendelinUser('test_user_%s' % generateRandomString(), None, 'ors/client')

    # Create two manager users: one associated to project_A and the second not associated to a project
    manager_user_A = self._createOrsWendelinUser('test_user_%s' % generateRandomString(), project_A_objects['project'].getRelativeUrl(), 'ors/manager')
    manager_user_N = self._createOrsWendelinUser('test_user_%s' % generateRandomString(), None, 'ors/manager')

    ors_N_reference = 'ors.test_%s' % generateRandomString()

    # Generate a Data Acquisition Unit without a project (should not happen in practice)
    data_acquisition_unit = self.portal.data_acquisition_unit_module.newContent(
      portal_type='Data Acquisition Unit',
      reference=ors_N_reference
    )
    data_acquisition_unit.validate()
    self.tic()
    data_supply = data_acquisition_unit.DataAcquisitionUnit_createORSDataSupply(batch=1)
    self.tic()

    # Force update on local roles to ensure proper permissions
    data_acquisition_unit.updateLocalRolesOnSecurityGroups()
    data_supply.updateLocalRolesOnSecurityGroups()

    # Perform an ingestion for the ORS not associated to a project
    ingestion_N_objects = self._getOrsIngestionObjects(self.test_ors_example_log_valid, ors_N_reference, stop_data_analysis=True)

    # Check that the client of project_A only has access to project_A documents
    client_user_A = project_A_objects['client_user']
    self._checkModulePermissions(client_user_A, False)
    self._checkIngestionDocumentsPermissions(client_user_A, ingestion_A_objects, False, True)
    self._checkIngestionDocumentsPermissions(client_user_A, ingestion_B_objects, False, False)
    self._checkIngestionDocumentsPermissions(client_user_A, ingestion_N_objects, False, False)

    # Check that the client of project_B only has access to project_B documents
    client_user_B = project_B_objects['client_user']
    self._checkModulePermissions(client_user_B, False)
    self._checkIngestionDocumentsPermissions(client_user_B, ingestion_A_objects, False, False)
    self._checkIngestionDocumentsPermissions(client_user_B, ingestion_B_objects, False, True)
    self._checkIngestionDocumentsPermissions(client_user_B, ingestion_N_objects, False, False)

    # Check that the client without a project does not have access to any document
    self._checkModulePermissions(client_user_N, False)
    self._checkIngestionDocumentsPermissions(client_user_N, ingestion_A_objects, False, False)
    self._checkIngestionDocumentsPermissions(client_user_N, ingestion_B_objects, False, False)
    self._checkIngestionDocumentsPermissions(client_user_N, ingestion_N_objects, False, False)

    # Check that both managers, whether assigned to a project, have access to all documents
    self._checkModulePermissions(manager_user_A, True)
    self._checkIngestionDocumentsPermissions(manager_user_A, ingestion_A_objects, True, True)
    self._checkIngestionDocumentsPermissions(manager_user_A, ingestion_B_objects, True, False)
    self._checkIngestionDocumentsPermissions(manager_user_A, ingestion_N_objects, True, False)

    self._checkModulePermissions(manager_user_N, True)
    self._checkIngestionDocumentsPermissions(manager_user_N, ingestion_A_objects, True, False)
    self._checkIngestionDocumentsPermissions(manager_user_N, ingestion_B_objects, True, False)
    self._checkIngestionDocumentsPermissions(manager_user_N, ingestion_N_objects, True, False)